import gettext
import logging
import os
import re

if os.name == "nt":
    import webbrowser
    import xml.etree.ElementTree

try:
    import gi
    gi.require_version('GdkPixbuf', '2.0')
    from gi.repository import GdkPixbuf
    GDK_PIXBUF_AVAILABLE = True
except (ValueError, ImportError):
    GDK_PIXBUF_AVAILABLE = False

try:
    import gi
    gi.require_version('Gdk', '3.0')
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gdk
    from gi.repository import Gtk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

import openpaperwork_core  # noqa: E402

from . import deps  # noqa: E402


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.cache = {}

    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_resources'
        ]

    def get_deps(self):
        return [
            {
                'interface': 'l10n_init',
                'defaults': [
                    'openpaperwork_core.l10n.python',
                    'openpaperwork_gtk.l10n',
                ],
            },
            {
                'interface': 'resources',
                'defaults': ['openpaperwork_core.resources.setuptools'],
            },
        ]

    def chkdeps(self, out: dict):
        if not GDK_PIXBUF_AVAILABLE:
            out['gdk_pixbuf'].update(deps.GDK_PIXBUF)
        if not GTK_AVAILABLE:
            out['gtk'].update(deps.GTK)

    @staticmethod
    def _translate_xml(xml_str):
        root = xml.etree.ElementTree.fromstring(xml_str)

        translation_domain = "openpaperwork_gtk"
        if 'domain' in root.attrib:
            translation_domain = root.attrib['domain']

        labels = root.findall('.//*[@translatable="yes"]')
        for label in labels:
            label.text = gettext.dgettext(translation_domain, label.text)

        out = xml.etree.ElementTree.tostring(root, encoding='UTF-8')
        return out.decode('utf-8')

    @staticmethod
    def _windows_fix_widgets(widget_tree):
        def open_uri(uri):

            # XXX(Jflesch): Seems we get some garbarge in the URI sometimes ?!
            uri = uri.replace("\u202f", "")

            LOGGER.info("Opening URI [%s]", uri)
            webbrowser.open(uri)

        for obj in widget_tree.get_objects():
            if isinstance(obj, Gtk.LinkButton):
                obj.connect(
                    "clicked",
                    lambda button: open_uri(button.get_uri())
                )
            elif (isinstance(obj, Gtk.AboutDialog) or
                    isinstance(obj, Gtk.Label)):
                obj.connect(
                    "activate-link",
                    lambda widget, uri: open_uri(uri)
                )

    def gtk_load_widget_tree(self, pkg, filename):
        """
        Load a .glade file

        Arguments:
            pkg -- Python package name
            filename -- css file name to load

        Returns:
            GTK Widget Tree
        """

        if not GTK_AVAILABLE:
            LOGGER.error("gtk_load_widget_tree(): GTK is not available")
            return None

        LOGGER.debug("Loading GTK widgets from %s:%s", pkg, filename)

        screen = Gdk.Screen.get_default()
        if screen is None:
            LOGGER.warning(
                "Cannot load widget tree: Gdk.Screen.get_default()"
                " returned None"
            )
            return None

        k = (pkg, filename)
        if k not in self.cache:
            try:
                filepath = self.core.call_success(
                    "resources_get_file", pkg, filename
                )
                with self.core.call_success("fs_open", filepath, 'r') as fd:
                    xml = fd.read()
                if os.name == "nt":
                    # WORKAROUND(Jflesch):
                    # for some reason,
                    # Gtk.Builder.new_from_file()/new_from_string doesn't
                    # translate on Windows
                    xml = self._translate_xml(xml)
                self.cache[k] = xml
            except Exception:
                LOGGER.error(
                    "Failed to load widget tree from file %s:%s",
                    pkg, filename
                )
                raise

        try:
            content = self.cache[k]
            widget_tree = Gtk.Builder.new_from_string(content, -1)
            if os.name == "nt":
                self._windows_fix_widgets(widget_tree)
            return widget_tree
        except Exception:
            LOGGER.error("Failed to load widget tree %s:%s", pkg, filename)
            raise

    def gtk_load_css(self, pkg, filename):
        """
        Load a .css file

        Arguments:
            pkg -- Python package name
            filename -- css file name to load.
        """
        if not GTK_AVAILABLE:
            LOGGER.error("gtk_load_css(): GTK is not available")
            return None

        LOGGER.debug("Loading CSS from %s:%s", pkg, filename)

        k = (pkg, filename)

        if k not in self.cache:
            try:
                filepath = self.core.call_success(
                    "resources_get_file", pkg, filename
                )
                with self.core.call_success("fs_open", filepath, 'rb') as fd:
                    self.cache[k] = fd.read()
            except Exception:
                LOGGER.error("Failed to load CSS file %s:%s", pkg, filename)
                raise

        try:
            content = self.cache[k]

            css_provider = Gtk.CssProvider()
            css_provider.load_from_data(content)

            screen = Gdk.Screen.get_default()
            if screen is None:
                LOGGER.warning(
                    "Cannot apply CSS: Gdk.Screen.get_default()"
                    " returned None"
                )
                return None

            Gtk.StyleContext.add_provider_for_screen(
                Gdk.Screen.get_default(), css_provider,
                Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
            )
        except Exception:
            LOGGER.error("Failed to load CSS %s:%s", pkg, filename)
            raise

        return True

    def gtk_load_pixbuf(self, pkg, filename):
        filepath = self.core.call_success("resources_get_file", pkg, filename)
        if filepath is None:
            return None
        filepath = self.core.call_success("fs_unsafe", filepath)
        return GdkPixbuf.Pixbuf.new_from_file(filepath)

    def gtk_fix_headerbar_buttons(self, headerbar):
        settings = Gtk.Settings.get_default()
        default_layout = settings.get_property("gtk-decoration-layout")
        default_layout = re.split("[:,]", default_layout)

        # disable the elements that are not enabled globally
        layout = headerbar.get_decoration_layout().split(":", 1)
        layout = [side.split(",") for side in layout]
        layout = ":".join([
            ",".join([
                element for element in side
                if element in default_layout
            ])
            for side in layout
        ])

        headerbar.set_decoration_layout(layout)



import openpaperwork_core.tests.local_file


PLUGIN_NAME = "openpaperwork_gtk.fs.gio"


class TestSafe(openpaperwork_core.tests.local_file.AbstractTestSafe):
    def get_plugin_name(self):
        return PLUGIN_NAME


class TestUnsafe(openpaperwork_core.tests.local_file.AbstractTestUnsafe):
    def get_plugin_name(self):
        return PLUGIN_NAME


class TestOpen(openpaperwork_core.tests.local_file.AbstractTestOpen):
    def get_plugin_name(self):
        return PLUGIN_NAME


class TestExists(openpaperwork_core.tests.local_file.AbstractTestExists):
    def get_plugin_name(self):
        return PLUGIN_NAME


class TestListDir(openpaperwork_core.tests.local_file.AbstractTestListDir):
    def get_plugin_name(self):
        return PLUGIN_NAME


class TestRename(openpaperwork_core.tests.local_file.AbstractTestRename):
    def get_plugin_name(self):
        return PLUGIN_NAME


class TestUnlink(openpaperwork_core.tests.local_file.AbstractTestUnlink):
    def get_plugin_name(self):
        return PLUGIN_NAME


class TestGetMtime(openpaperwork_core.tests.local_file.AbstractTestGetMtime):
    def get_plugin_name(self):
        return PLUGIN_NAME


class TestGetsize(openpaperwork_core.tests.local_file.AbstractTestGetsize):
    def get_plugin_name(self):
        return PLUGIN_NAME


class TestIsdir(openpaperwork_core.tests.local_file.AbstractTestIsdir):
    def get_plugin_name(self):
        return PLUGIN_NAME


class TestCopy(openpaperwork_core.tests.local_file.AbstractTestCopy):
    def get_plugin_name(self):
        return PLUGIN_NAME


class TestMkdirP(openpaperwork_core.tests.local_file.AbstractTestMkdirP):
    def get_plugin_name(self):
        return PLUGIN_NAME


class TestTemp(openpaperwork_core.tests.local_file.AbstractTestTemp):
    def get_plugin_name(self):
        return PLUGIN_NAME



GDK_PIXBUF = {
    'debian': 'gir1.2-gdkpixbuf-2.0',
    'linuxmint': 'gir1.2-gdkpixbuf-2.0',
    'raspbian': 'gir1.2-gdkpixbuf-2.0',
    'ubuntu': 'gir1.2-gdkpixbuf-2.0',
}

GTK = {
    'debian': 'gir1.2-gtk-3.0',
    'fedora': 'gtk3',
    'gentoo': 'x11-libs/gtk+',
    'linuxmint': 'gir1.2-gtk-3.0',
    'raspbian': 'gir1.2-gtk-3.0',
    'suse': 'python-gtk',
    'ubuntu': 'gir1.2-gtk-3.0',
}

NOTIFY = {
    'debian': 'gir1.2-notify-0.7',
    'fedora': 'libnotify',
    'ubuntu': 'gir1.2-notify-0.7',
}

PANGO = {
    'debian': 'gir1.2-pango-1.0',
    'linuxmint': 'gir1.2-pango-1.0',
    'raspbian': 'gir1.2-pango-1.0',
    'ubuntu': 'gir1.2-pango-1.0',
}



import io
import logging

import PIL
import PIL.Image

try:
    from gi.repository import GLib
    GLIB_AVAILABLE = True
except (ValueError, ImportError):
    GLIB_AVAILABLE = False

try:
    import gi
    gi.require_version('GdkPixbuf', '2.0')
    from gi.repository import GdkPixbuf
    GDK_PIXBUF_AVAILABLE = True
except (ValueError, ImportError):
    GDK_PIXBUF_AVAILABLE = False


import openpaperwork_core

from .. import deps


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return [
            'chkdeps',
            'pixbuf_pillow',
        ]

    def chkdeps(self, out: dict):
        if not GLIB_AVAILABLE:
            out['glib'].update(deps.GLIB)
        if not GDK_PIXBUF_AVAILABLE:
            out['gdk_pixbuf'].update(deps.GDK_PIXBUF)

    @staticmethod
    def pixbuf_to_pillow(pixbuf):
        (width, height) = (pixbuf.get_width(), pixbuf.get_height())
        pixels = pixbuf.get_pixels()
        colors = "RGB"
        if (width * height * 4 == len(pixels)):
            colors = "RGBA"
        return PIL.Image.frombytes(
            colors,
            (width, height),
            pixbuf.get_pixels()
        )

    def pillow_to_pixbuf(self, img):
        """
        Convert an image object to a GDK pixbuf
        """
        if not GDK_PIXBUF_AVAILABLE:
            return None

        if img is None:
            return None
        img = img.convert("RGB")

        if hasattr(GdkPixbuf.Pixbuf, 'new_from_bytes'):
            data = GLib.Bytes.new(img.tobytes())
            (width, height) = img.size
            return GdkPixbuf.Pixbuf.new_from_bytes(
                data, GdkPixbuf.Colorspace.RGB,
                False, 8, width, height, width * 3
            )

        file_desc = io.BytesIO()
        try:
            img.save(file_desc, "ppm")
            contents = file_desc.getvalue()
        finally:
            file_desc.close()
        loader = GdkPixbuf.PixbufLoader.new_with_type("pnm")
        try:
            loader.write(contents)
            pixbuf = loader.get_pixbuf()
        finally:
            loader.close()
        return pixbuf



import logging
import math

try:
    import cairo
    CAIRO_AVAILABLE = True
except (ImportError, ValueError):
    CAIRO_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.deps

from . import _


LOGGER = logging.getLogger(__name__)
SCREENSHOT_DATE_FORMAT = "%Y%m%d_%H%M_%S"
MAX_DAYS = 31
MAX_UNCAUGHT_EXCEPTION_SCREENSHOTS = 5


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.windows = []
        self.archiver = None
        self.nb_uncaught = 0

    def get_interfaces(self):
        return [
            'bug_report_attachments',
            'chkdeps',
            'gtk_window_listener',
            'uncaught_exception_listener',
            'screenshot',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'file_archives',
                'defaults': ['openpaperwork_core.archives'],
            },
            {
                'interface': 'fs',
                'defaults': [
                    'openpaperwork_core.fs.memory',
                    'openpaperwork_core.fs.python',
                ],
            },
            {
                'interface': 'uncaught_exception',
                'defaults': ['openpaperwork_core.uncaught_exception'],
            },
        ]

    def init(self, core):
        super().init(core)
        self.archiver = self.core.call_success(
            "file_archive_get", storage_name="screenshots",
            file_extension="png"
        )

    def chkdeps(self, out: dict):
        if not CAIRO_AVAILABLE:
            out['cairo'] = openpaperwork_core.deps.CAIRO

    def on_gtk_window_opened(self, window):
        self.windows.append(window)

    def on_gtk_window_closed(self, window):
        self.windows.remove(window)

    def _snap_screenshot(self, fd):
        image_size = (0, 0)

        for window in self.windows:
            if not window.is_drawable():
                LOGGER.warning("Window %s cannot be screenshoted", window)
            else:
                LOGGER.info("Screenshoting window %s", window)

        for window in self.windows:
            if not window.is_drawable():
                continue
            alloc = window.get_allocation()
            image_size = (
                max(image_size[0], alloc.width),
                image_size[1] + alloc.height
            )

        surface = cairo.ImageSurface(cairo.FORMAT_RGB24, *image_size)
        cairo_ctx = cairo.Context(surface)

        for window in self.windows:
            if not window.is_drawable():
                continue
            alloc = window.get_allocation()
            window.draw(cairo_ctx)
            cairo_ctx.translate(0, alloc.height)

        surface.write_to_png(fd)

    def screenshot_snap_all_promise(self, temporary=True, name="screenshots"):
        if temporary:
            (file_url, fd) = self.core.call_success(
                "fs_mktemp", prefix="{}_".format(name), suffix=".png",
                mode="wb", on_disk=True
            )
        else:
            file_url = self.archiver.get_new(name=name)
            fd = self.core.call_success("fs_open", file_url, mode='wb')

        promise = openpaperwork_core.promise.Promise(
            self.core, self.core.call_all, args=("on_screenshot_before",)
        )
        promise = promise.then(lambda *args, **kwargs: None)
        promise = promise.then(
            self.core.call_one, "mainloop_execute",
            self._snap_screenshot, fd
        )
        promise = promise.then(lambda *args, **kwargs: None)
        promise = promise.catch(lambda *args, **kwargs: fd.close())
        promise = promise.then(fd.close)
        promise = promise.then(lambda *args, **kwargs: None)
        promise = promise.then(self.core.call_all, "on_screenshot_after")
        promise = promise.then(lambda *args, **kwargs: None)

        return (file_url, promise)

    def screenshot_snap_widget(
                self, gtk_widget, out_file_url, margins=(20, 20, 20, 20),
                highlight=(0.67, 0.80, 0.93)
            ):
        assert(out_file_url.endswith(".png"))

        if not gtk_widget.is_drawable() or not gtk_widget.get_visible():
            LOGGER.warning(
                "%s is not visible. Cannot screenshot", gtk_widget
            )
            return None

        # find the GTK window of the widget
        window = gtk_widget.get_toplevel()
        if not window.is_drawable() or not window.get_visible():
            LOGGER.warning(
                "%s's window is not visible. Cannot screenshot",
                gtk_widget
            )
            return None

        # take a screenshot of the whole window
        win_alloc = window.get_allocation()
        win_surface = cairo.ImageSurface(
            cairo.FORMAT_RGB24, win_alloc.width, win_alloc.height
        )
        cairo_ctx = cairo.Context(win_surface)
        window.draw(cairo_ctx)

        widget_position = gtk_widget.translate_coordinates(window, 0, 0)
        widget_alloc = gtk_widget.get_allocation()

        # highlight
        if highlight is not None:
            cairo_ctx.save()
            try:
                cairo_ctx.new_path()
                border_width = 5
                margin = 5

                cairo_ctx.set_source_rgba(*highlight, 0.85)
                cairo_ctx.set_line_width(border_width)

                center = (
                    widget_position[0] + (widget_alloc.width / 2),
                    widget_position[1] + (widget_alloc.height / 2),
                )
                cairo_ctx.translate(*center)

                radius = (
                    (widget_alloc.width / 2) + (widget_alloc.height / 2)
                    + margin
                )
                if widget_alloc.width > widget_alloc.height:
                    scale = (1.0, widget_alloc.height / widget_alloc.width)
                else:
                    scale = (widget_alloc.width / widget_alloc.height, 1.0)

                cairo_ctx.scale(*scale)
                cairo_ctx.arc(0, 0, radius, 0, 2 * math.pi)
                cairo_ctx.scale(1.0 / scale[0], 1.0 / scale[1])
                cairo_ctx.stroke()
            finally:
                cairo_ctx.restore()

        # check it's visible enough
        start = (
            max(0, widget_position[0]),
            max(0, widget_position[1]),
        )
        size = (
            min(win_alloc.width - start[0], widget_alloc.width),
            min(win_alloc.height - start[1], widget_alloc.height),
        )
        if size[0] <= 15 or size[1] <= 15:
            LOGGER.warning(
                "%s is folded/hidden. Cannot screenshot", gtk_widget
            )
            return None

        # then cut it
        start = (
            max(0, widget_position[0] - margins[0]),
            max(0, widget_position[1] - margins[1])
        )
        size = (
            min(
                win_alloc.width - start[0],
                widget_alloc.width + margins[0] + margins[2]
            ),
            min(
                win_alloc.height - start[1],
                widget_alloc.height + margins[1] + margins[3],
            )
        )

        LOGGER.info(
            "Making screenshot of %s: %s (%s, %s)",
            gtk_widget, out_file_url, start, size
        )
        surface = cairo.ImageSurface(cairo.FORMAT_RGB24, *size)
        cairo_ctx = cairo.Context(surface)
        cairo_ctx.translate(-start[0], -start[1])
        cairo_ctx.set_source_surface(win_surface)
        cairo_ctx.paint()

        with self.core.call_success("fs_open", out_file_url, "wb") as fd:
            surface.write_to_png(fd)
        return True

    def on_uncaught_exception(self, exc_info):
        self.nb_uncaught += 1
        if self.nb_uncaught > MAX_UNCAUGHT_EXCEPTION_SCREENSHOTS:
            # limit the number of screenshots in one session.
            return

        LOGGER.info("Uncaught exception. Taking screenshots")
        (_, promise) = self.screenshot_snap_all_promise(
            name="uncaught_exception_screenshots"
        )
        promise.schedule()

    def bug_report_get_attachments(self, out: dict):
        for (date, file_url) in self.archiver.get_archived():
            out[file_url] = {
                'date': date,
                'include_by_default': False,
                'file_type': _("App. screenshots"),
                'file_url': file_url,
                'file_size': self.core.call_success("fs_getsize", file_url),
            }

        out['screenshot_now'] = {
            'include_by_default': False,
            'date': None,
            'file_type': _("App. screenshots"),
            'file_url': _("Select to generate"),
            'file_size': 0,
        }

    def _update_attachment(self, file_url, *args):
        self.core.call_all(
            "bug_report_update_attachment", "screenshot_now", {
                "file_url": file_url,
                "file_size": self.core.call_success("fs_getsize", file_url),
            }, *args
        )

    def on_bug_report_attachment_selected(self, attachment_id, *args):
        if attachment_id != 'screenshot_now':
            return
        (file_url, promise) = self.screenshot_snap_all_promise(temporary=True)
        promise = promise.then(self._update_attachment, file_url, *args)
        promise.schedule()



import logging
import math
import threading
import time

import openpaperwork_core
import openpaperwork_core.deps

try:
    import gi
    gi.require_version('Gdk', '3.0')
    from gi.repository import Gdk
    GDK_AVAILABLE = True
except (ImportError, ValueError):
    GDK_AVAILABLE = False

NEEDS_ATTENTION_TIMEOUT = 2.1
LOGGER = logging.getLogger(__name__)
TIME_BETWEEN_UPDATES = 0.3

# Tasks are often chained one after the other. We don't want the button/popover
# to disappear and reappear continually. So when a task ends, we
# give them some extra time to live.
# STAY_ALIVES is the number of updates we wait before hiding them.
STAY_ALIVES = int(2.0 / TIME_BETWEEN_UPDATES)


class ProgressWidget(object):
    def __init__(self, core):
        self.core = core
        self.widget = None

        # A thread updates the widgets every 300ms. We don't update them
        # each time on_progress() is called to not degrade performanes
        self.thread = None
        self.lock = threading.Lock()
        self.progress_widget_trees = {}
        # self.progresses is only used to transmist new progress updates
        # to the thread
        self.progresses = {}
        self.button_widget_tree = None
        self.details_widget_tree = None

        self.stay_alives = STAY_ALIVES

        self.button_widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "openpaperwork_gtk.widgets.progress",
            "progress_button.glade"
        )
        if self.button_widget_tree is None:
            # init must still work so 'chkdeps' is still available
            LOGGER.error("Failed to load widget tree")
            return
        self.details_widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "openpaperwork_gtk.widgets.progress",
            "progress_popover.glade"
        )
        if self.details_widget_tree is None:
            # init must still work so 'chkdeps' is still available
            LOGGER.error("Failed to load widget tree")
            return

        self.button_widget_tree.get_object("progress_button").set_popover(
            self.details_widget_tree.get_object(
                "progresses_popover"
            )
        )
        self.button_widget_tree.get_object("progress_icon").connect(
            "draw", self._on_icon_draw
        )

        self.widget = self.button_widget_tree.get_object("progress_revealer")

    def needs_attention(self):
        button = self.button_widget_tree.get_object("progress_button")
        button.get_style_context().add_class("progress_button_needs_attention")
        self.core.call_success(
            "mainloop_schedule",
            button.get_style_context().remove_class,
            "progress_button_needs_attention",
            delay_s=NEEDS_ATTENTION_TIMEOUT
        )

    def _thread(self):
        self.stay_alives = STAY_ALIVES
        while self.thread is not None:
            time.sleep(TIME_BETWEEN_UPDATES)
            self.core.call_one(
                "mainloop_execute", self._upd_progress_widgets
            )

    def _upd_progress_widgets(self):
        with self.lock:
            for (upd_type, (progress, description)) in self.progresses.items():
                self._upd_progress_widget(
                    upd_type, progress, description
                )
            self.progresses = {}

            if len(self.progress_widget_trees) > 0:
                self.stay_alives = STAY_ALIVES
                return

            if self.stay_alives > 0:
                self.stay_alives -= 1
                return

            self.button_widget_tree.get_object(
                "progress_revealer"
            ).set_reveal_child(False)
            self.thread = None
            return

    def _upd_progress_widget(self, upd_type, progress, description):
        if progress >= 1.0:  # deletion of progress
            if upd_type not in self.progress_widget_trees:
                LOGGER.warning(
                    "Got 2 notifications of end of task for '%s'",
                    upd_type
                )
                return
            LOGGER.info("Task '%s' has ended", upd_type)
            widget_tree = self.progress_widget_trees.pop(upd_type)
            box = self.details_widget_tree.get_object(
                "progresses_box"
            )
            details = widget_tree.get_object("progress_bar")
            box.remove(details)
            details.unparent()
            self.button_widget_tree.get_object("progress_button").queue_draw()
            self.needs_attention()

            LOGGER.info(
                "Task '%s' has ended (%d remaining)",
                upd_type, len(self.progress_widget_trees)
            )
            return

        if upd_type not in self.progress_widget_trees:  # creation of progress
            LOGGER.info(
                "Task '%s' has started (%d already active)",
                upd_type, len(self.progress_widget_trees)
            )
            widget_tree = self.core.call_success(
                "gtk_load_widget_tree",
                "openpaperwork_gtk.widgets.progress",
                "progress_details.glade"
            )
            box = self.details_widget_tree.get_object(
                "progresses_box"
            )
            box.add(widget_tree.get_object("progress_bar"))
            self.progress_widget_trees[upd_type] = widget_tree
            self.needs_attention()
        else:
            widget_tree = self.progress_widget_trees[upd_type]

        # update of progress
        progress_bar = widget_tree.get_object("progress_bar")
        progress_bar.set_fraction(progress)
        progress_bar.set_text(description if description is not None else "")

        self.button_widget_tree.get_object("progress_button").queue_draw()

        self.button_widget_tree.get_object(
            "progress_revealer"
        ).set_reveal_child(True)

    def on_progress(self, upd_type, progress, description=None):
        with self.lock:
            if progress > 1.0:
                LOGGER.warning(
                    "Invalid progression (%f) for [%s]",
                    progress, upd_type
                )
                progress = 1.0

            self.progresses[upd_type] = (progress, description)

            if self.thread is None:
                self.thread = threading.Thread(target=self._thread)
                self.thread.daemon = True
                self.thread.start()

    def _on_icon_draw(self, drawing_area, cairo_ctx):
        if len(self.progress_widget_trees) <= 0:
            ratio = 1.0
        else:
            ratio = sum([
                widget_tree.get_object("progress_bar").get_fraction()
                for widget_tree in self.progress_widget_trees.values()
            ]) / len(self.progress_widget_trees)

        # Translated in Python from Nautilus source code
        # (2020/02/29: src/nautilus-toolbar.c:on_operations_icon_draw())
        style_context = drawing_area.get_style_context()
        foreground = style_context.get_color(drawing_area.get_state_flags())
        background = foreground
        background.alpha *= 0.3

        w = drawing_area.get_allocated_width()
        h = drawing_area.get_allocated_height()

        Gdk.cairo_set_source_rgba(cairo_ctx, background)
        cairo_ctx.arc(w / 2.0, h / 2.0, min(w, h) / 2.0, 0, 2 * math.pi)
        cairo_ctx.fill()

        cairo_ctx.move_to(w / 2.0, h / 2.0)
        Gdk.cairo_set_source_rgba(cairo_ctx, foreground)
        cairo_ctx.arc(
            w / 2.0, h / 2.0, min(w, h) / 2.0, -math.pi / 2.0,
            (ratio * 2 * math.pi) - (math.pi / 2.0)
        )
        cairo_ctx.fill()


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.widgets = []

    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_progress_widget',
            'progress_listener',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_gtk.mainloop.glib'],
            },
        ]

    def init(self, core):
        super().init(core)
        self.core.call_success(
            "gtk_load_css",
            "openpaperwork_gtk.widgets.progress", "progress.css"
        )

    def chkdeps(self, out: dict):
        if not GDK_AVAILABLE:
            out['gdk'].update(openpaperwork_core.deps.GDK)

    def gtk_progress_make_widget(self):
        widget = ProgressWidget(self.core)
        if widget.widget is not None:
            # gtk may not be available
            self.widgets.append(widget)
        return widget.widget

    def on_progress(self, upd_type, progress, description=None):
        for widget in self.widgets:
            widget.on_progress(upd_type, progress, description)



import collections
import faulthandler
import logging
import threading

try:
    from gi.repository import GLib
    GLIB_AVAILABLE = True
except (ImportError, ValueError):
    GLIB_AVAILABLE = False


import openpaperwork_core

from .. import deps

LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 1000

    """
    A main loop based on GLib's mainloop.
    See `openpaperwork_core.mainloop.asyncio` for doc.
    """
    def __init__(self):
        super().__init__()
        self.halt_on_uncaught_exception = True
        self.log_uncaught = True
        self.loop = None
        self.loop_ident = None
        self.halt_cause = None
        self.task_count = 0

        self.lock = threading.RLock()
        self.active_tasks = collections.defaultdict(lambda: 0)

    def get_interfaces(self):
        return [
            "chkdeps",
            "mainloop",
        ]

    def chkdeps(self, out: dict):
        if not GLIB_AVAILABLE:
            out['glib'].update(deps.GLIB)

    def _check_mainloop_instantiated(self):
        if self.loop is None:
            self.loop = GLib.MainLoop.new(None, False)  # !running

    def mainloop(self, halt_on_uncaught_exception=True, log_uncaught=True):
        self._check_mainloop_instantiated()
        self.log_uncaught = log_uncaught
        self.halt_on_uncaught_exception = halt_on_uncaught_exception

        self.loop_ident = threading.current_thread().ident

        self.mainloop_schedule(self.core.call_all, "on_mainloop_start")

        try:
            self.loop.run()
        except Exception:
            faulthandler.dump_traceback()
            raise
        finally:
            self.loop_ident = None

        self.core.call_all("on_mainloop_quit")

        if self.halt_cause is not None:
            halt_cause = self.halt_cause
            self.halt_cause = None
            LOGGER.error("Main loop stopped because %s", str(halt_cause))
            raise halt_cause

        self.loop = None
        return True

    def mainloop_get_thread_id(self):
        return self.loop_ident

    def mainloop_quit_graceful(self):
        if self.loop is None:
            return None
        self.mainloop_schedule(self._mainloop_quit_graceful)
        return True

    def _mainloop_quit_graceful(self):
        quit_now = True

        with self.lock:
            # keep in mind this function is in a task too
            if self.task_count > 1:
                quit_now = False
                LOGGER.info(
                    "Quit graceful: Remaining tasks: %d", self.task_count - 1
                )
                for (k, v) in self.active_tasks.items():
                    LOGGER.info("Quit graceful: Remaining: %s = %d", k, v)

        if not quit_now:
            self.mainloop_schedule(
                self._mainloop_quit_graceful, delay_s=0.2
            )
            return

        LOGGER.info("Quit graceful: Quitting")

        self.mainloop_quit_now()

        with self.lock:
            self.task_count = 1  # we are actually the one task still running
            self.active_tasks = collections.defaultdict(lambda: 0)

    def mainloop_quit_now(self):
        if self.loop is None:
            return None

        with self.lock:
            self.loop.quit()
            self.loop = None
            self.task_count = 0
            self.active_tasks = collections.defaultdict(lambda: 0)

    def mainloop_ref(self, obj):
        with self.lock:
            self.task_count += 1
            self.active_tasks[str(obj)] += 1

    def mainloop_unref(self, obj):
        with self.lock:
            self.task_count -= 1
            assert(self.task_count >= 0)
            try:
                s = str(obj)
                self.active_tasks[s] -= 1
                if self.active_tasks[s] <= 0:
                    self.active_tasks.pop(s)
            except KeyError:
                pass

    def mainloop_schedule(self, func, *args, delay_s=0, **kwargs):
        assert(hasattr(func, '__call__'))

        with self.lock:
            self._check_mainloop_instantiated()

            self.task_count += 1
            self.active_tasks[str(func)] += 1

        def decorator(func, args):
            (args, kwargs) = args
            try:
                func(*args, **kwargs)
            except Exception as exc:
                if self.halt_on_uncaught_exception:
                    LOGGER.error(
                        "Main loop: uncaught exception (%s) ! Quitting",
                        func, exc_info=exc
                    )
                    self.halt_cause = exc
                    self.mainloop_quit_now()
                elif self.log_uncaught:
                    LOGGER.error(
                        "Main loop: uncaught exception (%s) !",
                        func, exc_info=exc
                    )
            finally:
                with self.lock:
                    self.task_count -= 1
                    try:
                        s = str(func)
                        self.active_tasks[s] -= 1
                        if self.active_tasks[s] <= 0:
                            self.active_tasks.pop(s)
                    except KeyError:
                        pass
            return False

        args = (args, kwargs)

        if delay_s is None:
            GLib.idle_add(decorator, func, args, priority=GLib.PRIORITY_LOW)
        else:
            GLib.timeout_add(delay_s * 1000, decorator, func, args)
        return True

    def mainloop_execute(self, func, *args, **kwargs):
        current = threading.current_thread().ident

        # XXX(Jflesch):
        # if self.loop_ident is None, it means the mainloop hasn't been started
        # yet --> we cannot run the function on the mainloop anyway, so
        # we assume we are on the same thread that will later run the main
        # loop.
        if self.loop_ident is None or current == self.loop_ident:
            return func(*args, **kwargs)

        event = threading.Event()
        out = [None]
        exc = [None]

        def get_result():
            try:
                out[0] = func(*args, **kwargs)
            except Exception as e:
                LOGGER.warning(
                    "mainloop_execute exception (func=%s, args=%s, kwargs=%s)",
                    func, args, kwargs, exc_info=e
                )
                exc[0] = e
            event.set()

        self.mainloop_schedule(get_result)
        event.wait()

        if exc[0] is not None:
            raise exc[0]
        return out[0]



import collections
import logging
import math
import random
import statistics

import openpaperwork_core
import openpaperwork_core.deps
import openpaperwork_gtk.deps

from openpaperwork_gtk import _


GI_AVAILABLE = False
GTK_AVAILABLE = False
PANGO_AVAILABLE = False

try:
    import gi
    from gi.repository import GObject
    GI_AVAILABLE = True
except (ImportError, ValueError):
    pass

if GI_AVAILABLE:
    try:
        gi.require_version('Gdk', '3.0')
        gi.require_version('Gtk', '3.0')
        from gi.repository import Gdk
        from gi.repository import Gtk
        GTK_AVAILABLE = True
    except (ImportError, ValueError):
        pass

    try:
        gi.require_version('Pango', '1.0')
        gi.require_version('PangoCairo', '1.0')
        from gi.repository import Pango
        from gi.repository import PangoCairo
        PANGO_AVAILABLE = True
    except (ImportError, ValueError):
        pass


LOGGER = logging.getLogger(__name__)
SUM_COLOR = (0, 0, 0)


class Schema(object):
    def __init__(
            self,
            column_value_id_idx,
            column_line_id_idx,
            column_axis_x_values_idx,
            column_axis_x_names_idx,
            column_axis_y_values_idx,
            column_axis_y_names_idx,
            highlight_x_range=(math.inf, -math.inf)):
        self.column_value_id_idx = column_value_id_idx
        self.column_line_id_idx = column_line_id_idx
        self.column_axis_x_values_idx = column_axis_x_values_idx
        self.column_axis_x_names_idx = column_axis_x_names_idx
        self.column_axis_y_values_idx = column_axis_y_values_idx
        self.column_axis_y_names_idx = column_axis_y_names_idx
        self.highlight_x_range = highlight_x_range


class ColorGenerator(object):
    def __init__(self):
        self.colors = []
        self.color_idx = -1

    def reset(self):
        self.color_idx = -1

    def __next__(self):
        self.color_idx += 1
        if self.color_idx >= len(self.colors):
            self.colors.append((
                random.randint(0, 0xFF) / 0xFF,
                random.randint(0, 0xFF) / 0xFF,
                random.randint(0, 0xFF) / 0xFF,
            ))
        return self.colors[self.color_idx]


class DrawContext(object):
    """
    Custom drawing context around cairo context, so we can change
    the scale without changing the line width.
    """
    def __init__(self, ctx):
        self.ctx = ctx

    def get_cairo(self):
        if hasattr(self.ctx, 'get_cairo'):
            return self.ctx.get_cairo()
        return self.ctx

    def translate_pt(self, pt_x, pt_y):
        return (pt_x, pt_y)

    def untranslate_pt(self, pt_x, pt_y):
        return (pt_x, pt_y)

    def distance(self, pt_a_x, pt_a_y, pt_b_x, pt_b_y):
        d_x = abs(pt_a_x - pt_b_x)
        d_y = abs(pt_a_y - pt_b_y)
        return (d_x ** 2) + (d_y ** 2)

    def arc(self, x, y, radius, angle1, angle2):
        self.ctx.arc(x, y, radius, angle1, angle2)

    def save(self):
        self.ctx.save()

    def restore(self):
        self.ctx.restore()

    def set_source_rgb(self, r, g, b):
        self.ctx.set_source_rgb(r, g, b)

    def set_source_rgba(self, r, g, b, a):
        self.ctx.set_source_rgba(r, g, b, a)

    def set_line_width(self, line_width):
        self.ctx.set_line_width(line_width)

    def move_to(self, pt_x, pt_y):
        self.ctx.move_to(pt_x, pt_y)

    def line_to(self, pt_x, pt_y):
        self.ctx.line_to(pt_x, pt_y)

    def rectangle(self, x, y, w, h):
        self.ctx.rectangle(x, y, w, h)

    def stroke(self):
        self.ctx.stroke()

    def fill(self):
        self.ctx.fill()

    def translate(self, x, y):
        return DrawContextTranslated(self, x, y)

    def scale(self, x, y):
        return DrawContextScaled(self, x, y)


class DrawContextTranslated(DrawContext):
    def __init__(self, ctx, translation_x, translation_y):
        super().__init__(ctx)
        self.x = translation_x
        self.y = translation_y

    def translate_pt(self, pt_x, pt_y):
        (pt_x, pt_y) = (pt_x + self.x, pt_y + self.y)
        return self.ctx.translate_pt(pt_x, pt_y)

    def untranslate_pt(self, pt_x, pt_y):
        (pt_x, pt_y) = self.ctx.untranslate_pt(pt_x, pt_y)
        return (pt_x - self.x, pt_y - self.y)

    def distance(
            self,
            pt_untranslated_x, pt_untranslated_y,
            pt_translated_x, pt_translated_y):
        pt_a_x = pt_untranslated_x + self.x
        pt_a_y = pt_untranslated_y + self.y
        return self.ctx.distance(
            pt_a_x, pt_a_y,
            pt_translated_x, pt_translated_y,
        )

    def arc(self, x, y, radius, angle1, angle2):
        self.ctx.arc(x + self.x, y + self.y, radius, angle1, angle2)

    def move_to(self, pt_x, pt_y):
        self.ctx.move_to(pt_x + self.x, pt_y + self.y)

    def line_to(self, pt_x, pt_y):
        self.ctx.line_to(pt_x + self.x, pt_y + self.y)

    def rectangle(self, x, y, w, h):
        self.ctx.rectangle(x + self.x, y + self.y, w, h)


class DrawContextScaled(DrawContext):
    def __init__(self, ctx, scale_x, scale_y):
        super().__init__(ctx)
        self.x = scale_x
        self.y = scale_y

    def translate_pt(self, pt_x, pt_y):
        (pt_x, pt_y) = (pt_x * self.x, pt_y * self.y)
        return self.ctx.translate_pt(pt_x, pt_y)

    def untranslate_pt(self, pt_x, pt_y):
        (pt_x, pt_y) = self.ctx.untranslate_pt(pt_x, pt_y)
        return (pt_x / self.x, pt_y / self.y)

    def distance(
            self,
            pt_unscaled_x, pt_unscaled_y,
            pt_scaled_x, pt_scaled_y):
        pt_a_x = pt_unscaled_x * self.x
        pt_a_y = pt_unscaled_y * self.y
        return self.ctx.distance(
            pt_a_x, pt_a_y,
            pt_scaled_x, pt_scaled_y,
        )

    def arc(self, x, y, radius, angle1, angle2):
        self.ctx.arc(x * self.x, y * self.y, radius, angle1, angle2)

    def move_to(self, pt_x, pt_y):
        self.ctx.move_to(pt_x * self.x, pt_y * self.y)

    def line_to(self, pt_x, pt_y):
        self.ctx.line_to(pt_x * self.x, pt_y * self.y)

    def rectangle(self, x, y, w, h):
        self.ctx.rectangle(x * self.x, y * self.y, w * self.x, h * self.y)


class Point(object):
    RADIUS = 2
    LABEL_LINE_HEIGHT = 14

    def __init__(self, liststore_idx, x, y, label_x, label_y, color):
        self.liststore_idx = liststore_idx
        self.x = x
        self.y = y
        self.label_x = label_x
        self.label_y = label_y
        self.color = color

    def copy(self):
        return Point(
            self.liststore_idx,
            self.x, self.y,
            self.label_x, self.label_y,
            self.color
        )

    def __lt__(self, other):
        if self.x < other.x:
            return True
        if self.y < other.y:
            return True
        return False

    @staticmethod
    def merge(pts):
        merged_pt = pts[0].copy()
        merged_pt.y = statistics.mean([p.y for p in pts])
        merged_pt.label_y = "\n".join([p.label_y for p in pts])
        return merged_pt

    @staticmethod
    def merge_same_x(all_pts):
        pack = []
        last_x = None
        pts = []
        for pt in all_pts:
            if last_x is None:
                pack.append(pt)
                last_x = pt.x
            elif last_x == pt.x:
                pack.append(pt)
            else:
                pts.append(Point.merge(pack))
                last_x = pt.x
                pack = [pt]
        if len(pack) > 0:
            pts.append(Point.merge(pack))
        return pts

    @staticmethod
    def from_liststore_line(liststore_idx, line, schema, color):
        return Point(
            liststore_idx,
            line[schema.column_axis_x_values_idx],
            line[schema.column_axis_y_values_idx],
            line[schema.column_axis_x_names_idx],
            line[schema.column_axis_y_names_idx],
            color
        )

    def draw_label_x(self, widget_size, draw_ctx):
        height = (self.label_x.count("\n") + 1) * self.LABEL_LINE_HEIGHT

        cairo_ctx = draw_ctx.get_cairo()
        x = draw_ctx.translate_pt(self.x, widget_size[1] - self.y)[0]
        widget_size = draw_ctx.translate_pt(widget_size[0], widget_size[1])

        layout = PangoCairo.create_layout(cairo_ctx)
        layout.set_text(self.label_x, -1)

        txt_size = layout.get_size()
        if 0 in txt_size:
            return

        cairo_ctx.save()
        try:
            txt_scale = height / txt_size[1]
            txt_width = txt_size[0] * txt_scale
            if x <= widget_size[0] / 2:
                x += 5
            else:
                x -= (txt_width + 10)

            cairo_ctx.set_source_rgb(0, 0, 0)
            cairo_ctx.translate(x, 0)
            cairo_ctx.scale(txt_scale * Pango.SCALE, txt_scale * Pango.SCALE)
            PangoCairo.update_layout(cairo_ctx, layout)
            PangoCairo.show_layout(cairo_ctx, layout)
        finally:
            cairo_ctx.restore()

    def draw_label_y(self, widget_size, draw_ctx):
        height = (self.label_y.count("\n") + 1) * self.LABEL_LINE_HEIGHT

        cairo_ctx = draw_ctx.get_cairo()
        y = draw_ctx.translate_pt(self.x, widget_size[1] - self.y)[1]
        widget_size = draw_ctx.translate_pt(widget_size[0], widget_size[1])

        layout = PangoCairo.create_layout(cairo_ctx)
        layout.set_text(self.label_y, -1)

        txt_size = layout.get_size()
        if 0 in txt_size:
            return

        cairo_ctx.save()
        try:
            txt_scale = height / txt_size[1]
            if y <= widget_size[1] / 2:
                y += 2
            else:
                y -= (height + 2)
            cairo_ctx.set_source_rgb(0, 0, 0)
            cairo_ctx.translate(0, y)
            cairo_ctx.scale(txt_scale * Pango.SCALE, txt_scale * Pango.SCALE)
            PangoCairo.update_layout(cairo_ctx, layout)
            PangoCairo.show_layout(cairo_ctx, layout)
        finally:
            cairo_ctx.restore()

    def draw_highlight(self, widget_size, draw_ctx):
        draw_ctx.set_source_rgb(0.57, 0.7, 0.837)
        draw_ctx.set_line_width(1)

        # TODO(Jflesch): we should use the real widget size, and not
        # the widget size relative to point (0, 0)
        draw_ctx.move_to(-widget_size[0] * 2, widget_size[1] - self.y)
        draw_ctx.line_to(widget_size[0] * 2, widget_size[1] - self.y)
        draw_ctx.stroke()

        # TODO(Jflesch): we should use the real widget size, and not
        # the widget size relative to point (0, 0)
        draw_ctx.move_to(self.x, -widget_size[1] * 2)
        draw_ctx.line_to(self.x, widget_size[1] * 2)
        draw_ctx.stroke()

        self.draw_label_x(widget_size, draw_ctx)
        self.draw_label_y(widget_size, draw_ctx)

    def distance(self, widget_size, draw_ctx, x, y):
        return draw_ctx.distance(self.x, widget_size[1] - self.y, x, y)


class Line(object):
    LEGEND_CIRCLE_RADIUS = 8
    LEGEND_LINE_HEIGHT = 16
    LEGEND_SPACING = 8

    def __init__(self, line_name, points, color, line_width=1.0):
        self.line_name = line_name
        self.points = points
        self.points.sort(key=lambda pt: pt.x)
        self.color = color
        self.minmax = ((math.inf, 0), (0, 0))
        self.line_width = line_width
        for pt in points:
            self.minmax = (
                (
                    min(self.minmax[0][0], pt.x),
                    max(self.minmax[0][1], pt.x),
                ),
                (
                    min(self.minmax[1][0], pt.y),
                    max(self.minmax[1][1], pt.y),
                ),
            )

    @staticmethod
    def from_liststore_lines(line_name, liststore_lines, schema, color):
        pts = [
            Point.from_liststore_line(liststore_idx, line, schema, color)
            for (liststore_idx, line) in liststore_lines
        ]
        pts = Point.merge_same_x(pts)
        return Line(line_name, pts, color)

    @staticmethod
    def make_line_sum(other_lines):
        pts = []
        for line in other_lines:
            pts += [
                (line.line_name, pt.copy())
                for pt in line.points
            ]
        pts.sort(key=lambda pt: pt[1].x)

        total_txt = _("Total: {}")

        last_y = collections.defaultdict(lambda: 0)
        for (line_name, pt) in pts:
            last_y[line_name] = pt.y
            pt.y = sum(last_y.values())
            pt.label_y = total_txt.format(pt.y)

        pts = [pt for (line_name, pt) in pts]
        pts = Point.merge_same_x(pts)
        return Line(_("Total"), pts, SUM_COLOR, 2.0)

    def draw_chart(self, widget_size, draw_ctx):
        draw_ctx.save()
        try:
            draw_ctx.set_line_width(self.line_width)
            draw_ctx.set_source_rgb(*self.color)
            for (idx, pt) in enumerate(self.points):
                if idx == 0:
                    draw_ctx.move_to(pt.x, widget_size[1] - pt.y)
                else:
                    draw_ctx.line_to(pt.x, widget_size[1] - pt.y)
            draw_ctx.stroke()
        finally:
            draw_ctx.restore()

    def draw_legend(self, widget_size, cairo_ctx, x, y):
        layout = PangoCairo.create_layout(cairo_ctx)
        layout.set_text(self.line_name, -1)
        txt_size = layout.get_size()
        if 0 in txt_size:
            return (x, y)

        cairo_ctx.save()
        try:
            txt_scale = self.LEGEND_LINE_HEIGHT / txt_size[1]

            txt_width = txt_size[0] * txt_scale
            width = txt_width
            width += 2 * self.LEGEND_CIRCLE_RADIUS
            width += 2 * self.LEGEND_SPACING
            if x + width >= widget_size[0]:
                y += self.LEGEND_LINE_HEIGHT + self.LEGEND_SPACING
                x = 0

            cairo_ctx.set_source_rgb(*self.color)
            cairo_ctx.arc(
                x + self.LEGEND_CIRCLE_RADIUS,
                y + self.LEGEND_CIRCLE_RADIUS,
                self.LEGEND_CIRCLE_RADIUS,
                0, math.pi * 2
            )
            cairo_ctx.fill()
            x += (2 * self.LEGEND_CIRCLE_RADIUS) + self.LEGEND_SPACING

            cairo_ctx.set_source_rgb(0.0, 0.0, 0.0)
            cairo_ctx.translate(x, y)
            cairo_ctx.scale(txt_scale * Pango.SCALE, txt_scale * Pango.SCALE)
            PangoCairo.update_layout(cairo_ctx, layout)
            PangoCairo.show_layout(cairo_ctx, layout)

            x += txt_width + self.LEGEND_SPACING
        finally:
            cairo_ctx.restore()
        return (x, y)

    def get_closest_point(self, widget_size, draw_ctx, x, y):
        return min((
            (point.distance(widget_size, draw_ctx, x, y), point)
            for point in self.points
        ), default=None)


class Lines(object):
    HIGHLIGHT_COLOR = (0.5, 0.5, 0.5, 0.5)

    def __init__(self, schema, liststore, color_generator):
        self.schema = schema
        self.liststore = liststore
        self.color_generator = color_generator
        self.active_point = None

        self.lines = []
        self.minmax = ((math.inf, 0), (0, 0),)

    def reset(self):
        self.lines = []

    def reload(self):
        self.minmax = ((math.inf, 0), (0, 0),)
        self.color_generator.reset()
        lines = collections.defaultdict(list)
        for (liststore_idx, line) in enumerate(self.liststore):
            line_name = line[self.schema.column_line_id_idx]
            lines[line_name].append((liststore_idx, line))
        self.lines = [
            Line.from_liststore_lines(
                k, v, self.schema, next(self.color_generator)
            )
            for (k, v) in lines.items()
        ]
        if len(self.lines) > 1:
            self.lines = [Line.make_line_sum(self.lines)] + self.lines
        for line in self.lines:
            self.minmax = (
                (
                    min(self.minmax[0][0], line.minmax[0][0]),
                    max(self.minmax[0][1], line.minmax[0][1]),
                ),
                (
                    min(self.minmax[1][0], line.minmax[1][0]),
                    max(self.minmax[1][1], line.minmax[1][1]),
                ),
            )

    def draw_chart_highlight(self, widget_size, draw_ctx):
        x_range = self.schema.highlight_x_range
        if x_range[1] <= x_range[0]:
            return
        if x_range[0] is -math.inf:
            x_range = (self.minmax[0][0], self.x_range[1])
        if x_range[1] is math.inf:
            x_range = (x_range[0], self.minmax[0][1])
        x_range = (
            (
                x_range[0] - self.minmax[0][0]
            ) / (
                self.minmax[0][1] - self.minmax[0][0]
            ),
            (
                x_range[1] - self.minmax[0][0]
            ) / (
                self.minmax[0][1] - self.minmax[0][0]
            ),
        )
        x_range = (x_range[0] * widget_size[0], x_range[1] * widget_size[0])

        draw_ctx.save()
        try:
            draw_ctx.set_source_rgba(*self.HIGHLIGHT_COLOR)
            draw_ctx.rectangle(
                int(x_range[0]) + 1, 0, x_range[1] - x_range[0], widget_size[1]
            )
            draw_ctx.fill()
        finally:
            draw_ctx.restore()

    def draw_chart(self, widget_size, draw_ctx):
        w = self.minmax[0][1] - self.minmax[0][0]
        h = self.minmax[1][1] - self.minmax[1][0]

        if w == 0 or h == 0:
            return

        draw_ctx = draw_ctx.scale(widget_size[0] / w, widget_size[1] / h)
        widget_size = (w, h)

        self.draw_chart_highlight(widget_size, draw_ctx)

        draw_ctx = draw_ctx.translate(
            -self.minmax[0][0], min(0, self.minmax[1][0])
        )
        widget_size = (widget_size[0] + self.minmax[0][0], widget_size[1])

        draw_ctx.save()
        try:
            for line in self.lines:
                line.draw_chart(widget_size, draw_ctx)
        finally:
            draw_ctx.restore()
        if self.active_point is not None:
            self.active_point.draw_highlight(widget_size, draw_ctx)

    def draw_legend(self, widget_size, cairo_ctx):
        x = 0
        y = 0
        for line in self.lines:
            (x, y) = line.draw_legend(widget_size, cairo_ctx, x, y)
        return (x, y)

    def get_closest_point(self, widget_size, draw_ctx, x, y):
        w = self.minmax[0][1] - self.minmax[0][0]
        h = self.minmax[1][1] - self.minmax[1][0]

        if w == 0 or h == 0:
            return None

        draw_ctx = draw_ctx.scale(widget_size[0] / w, widget_size[1] / h)
        widget_size = (w, h)

        draw_ctx = draw_ctx.translate(
            -self.minmax[0][0], min(0, self.minmax[1][0])
        )
        widget_size = (widget_size[0] - self.minmax[0][0], widget_size[1])

        return min((
            line.get_closest_point(widget_size, draw_ctx, x, y)
            for line in self.lines
        ), default=None)


class ChartDrawer(object):
    MARGIN = 5

    def __init__(self, core, schema, liststore, color_generator):
        self.core = core
        self.lines = Lines(schema, liststore, color_generator)

        self.chart_widget = Gtk.DrawingArea.new()
        self.chart_widget.set_size_request(-1, 200)
        self.chart_widget.set_visible(True)
        self.chart_widget.connect("draw", self.draw_chart)
        self.chart_widget.connect("realize", self._on_chart_realized)
        self.chart_widget.connect("motion-notify-event", self._on_mouse_motion)

        self.legend_widget = Gtk.DrawingArea.new()
        self.legend_widget.set_visible(True)
        self.legend_widget.connect("draw", self.draw_legend)

        liststore.connect("row-changed", self.reload)
        liststore.connect("row-deleted", self.reload)
        liststore.connect("row-inserted", self.reload)
        liststore.connect("rows-reordered", self.reload)

        self._on_chart_realized()

        self.has_changed = False
        self.reload_planned = False
        self._reload()

    def _on_chart_realized(self, *args, **kwargs):
        mask = Gdk.EventMask.POINTER_MOTION_MASK
        self.chart_widget.add_events(mask)
        if self.chart_widget.get_window() is not None:
            self.chart_widget.get_window().set_events(
                self.chart_widget.get_window().get_events() | mask
            )

    def reload(self, *args, **kwargs):
        self.lines.reset()
        self.chart_widget.queue_draw()
        self.legend_widget.queue_draw()
        if self.reload_planned:
            self.has_changed = True
        else:
            self.has_changed = False
            self.reload_planned = True
            promise = openpaperwork_core.promise.DelayPromise(
                self.core, delay_s=0.25
            )
            promise.then(self._reload)
            promise.schedule()

    def _reload(self):
        self.reload_planned = False
        if self.has_changed:
            self.reload()
            return
        self.lines.reload()
        self.chart_widget.queue_draw()
        self.legend_widget.queue_draw()

    def _on_mouse_motion(self, widget, event):
        draw_ctx = DrawContext(None)
        draw_ctx = draw_ctx.translate(self.MARGIN, self.MARGIN)
        widget_size = (
            widget.get_allocated_width() - (2 * self.MARGIN),
            widget.get_allocated_height() - (2 * self.MARGIN),
        )
        r = self.lines.get_closest_point(
            widget_size, draw_ctx, event.x, event.y
        )
        if r is None:
            self.lines.active_point = None
            return
        (dist, point) = r
        if self.lines.active_point is not point:
            self.lines.active_point = point
            self.chart_widget.queue_draw()

    def draw_chart(self, widget, cairo_ctx):
        widget_size = (
            widget.get_allocated_width() - (2 * self.MARGIN),
            widget.get_allocated_height() - (2 * self.MARGIN),
        )
        draw_ctx = DrawContext(cairo_ctx)
        draw_ctx = draw_ctx.translate(self.MARGIN, self.MARGIN)
        self.lines.draw_chart(widget_size, draw_ctx)

    def draw_legend(self, widget, cairo_ctx):
        widget_size = (
            widget.get_allocated_width() - (2 * self.MARGIN),
            widget.get_allocated_height() - (2 * self.MARGIN),
        )
        cairo_ctx.translate(self.MARGIN, self.MARGIN)
        (x, y) = self.lines.draw_legend(widget_size, cairo_ctx)
        if x > 0:
            # the y we got is relative to the top of the current line.
            # the height is starting at the bottom of the current line.
            y += Line.LEGEND_SPACING + Line.LEGEND_LINE_HEIGHT
        if y + self.MARGIN != widget_size[1]:
            widget.set_size_request(-1, y + self.MARGIN)


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.color_generator = ColorGenerator()

    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_charts_lines',
        ]

    def chkdeps(self, out: dict):
        if not GTK_AVAILABLE:
            out['gtk'].update(openpaperwork_gtk.deps.GTK)
        if not PANGO_AVAILABLE:
            out['pango'].update(openpaperwork_core.deps.PANGO)

    def gtk_charts_lines_get(
            self, liststore, *args, **kwargs):
        """
        Return a Gtk.Widget showing the specified line chart.

        Can display multiple lines on single chart.
        In the model, each row is a value (x, y) for a given line.

        Arguments:
        - model: Gtk.ListStore to use as model
        - column_value_id_idx: column in the model that contains the value IDs.
        - column_line_id_idx: column that contains the name of the line
          to which this value belongs.
        - column_axis_x_values_idx: column that contains the values for the X
          axis. Must contain integers, floats or doubles.
        - column_axis_x_names_idx: column that contains the names corresponding
          to the values on the X axis. Must be strings.
        - column_axis_y_values_idx: column that contains the values for the Y
          axis. Must contain integers, floats or doubles.
        - column_axis_y_names_idx: column that contains the names corresponding
          toi the values on the Y axis. Must be strings
        """
        schema = Schema(*args, **kwargs)
        return ChartDrawer(self.core, schema, liststore, self.color_generator)


if __name__ == "__main__":
    model = Gtk.ListStore.new((
        GObject.TYPE_STRING,  # value id
        GObject.TYPE_STRING,  # line id
        GObject.TYPE_INT64,  # value X
        GObject.TYPE_INT64,  # value Y
        GObject.TYPE_STRING,  # name X
        GObject.TYPE_STRING,  # name Y
    ))
    model_content = (
        ("value A55", "line A", 5, 5, "A5", "A5"),
        ("value B55", "line B", 5, 5, "B5", "B5"),
        ("value A67", "line A", 6, 7, "A6", "A7"),
        ("value B66", "line B", 6, 6, "B6", "B6"),
        # graph must support having many values at the same position in X
        ("value A77", "line A", 7, 7, "A7", "A7"),
        ("value A73", "line A", 7, 3, "A7", "A3"),
        ("value A79", "line A", 7, 9, "A7", "A9"),
        ("value B79", "line B", 7, 9, "B7", "B9"),
        ("value A83", "line A", 8, 3, "A8", "A3"),
        ("value B88", "line B", 8, 8, "B8", "B8"),
        ("value A97", "line A", 9, 7, "A9", "A7"),
        ("value B95", "line B", 9, 5, "B9", "B5"),
        ("value A10-5", "line A", 10, -5, "A10", "A-5"),
        ("value B910", "line B", 9, 10, "B9", "B10"),
    )
    model.clear()
    for line in model_content:
        model.append(line)

    core = openpaperwork_core.Core()
    core.load("openpaperwork_gtk.widgets.charts.lines")
    core.init()

    chart_lines = core.call_success(
        "gtk_charts_lines_get", model,
        column_value_id_idx=0,
        column_line_id_idx=1,
        column_axis_x_values_idx=2,
        column_axis_y_values_idx=3,
        column_axis_x_names_idx=4,
        column_axis_y_names_idx=5
    )

    window = Gtk.Window()
    window.set_default_size(600, 600)
    window.add(chart_lines.chart_widget)
    window.show_all()

    window_legend = Gtk.Window()
    window_legend.set_default_size(100, 100)
    window_legend.add(chart_lines.legend_widget)
    window_legend.show_all()

    Gtk.main()



import logging

try:
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.promise
import openpaperwork_gtk.deps


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    """
    Provides a simple way to show a Yes/No question popup (usually something
    along the line of "are you really really really sure you want to do
    that ?").
    """

    def __init__(self):
        super().__init__()
        self.windows = []

    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_dialog_yes_no',
            'gtk_window_listener',
        ]

    def chkdeps(self, out: dict):
        if not GTK_AVAILABLE:
            out['gtk'].update(openpaperwork_gtk.deps.GTK)

    def on_gtk_window_opened(self, window):
        self.windows.append(window)

    def on_gtk_window_closed(self, window):
        self.windows.remove(window)

    def gtk_show_dialog_yes_no(self, origin, msg, *args, **kwargs):
        confirm = Gtk.MessageDialog(
            parent=self.windows[-1],
            flags=(
                Gtk.DialogFlags.MODAL |
                Gtk.DialogFlags.DESTROY_WITH_PARENT
            ),
            message_type=Gtk.MessageType.WARNING,
            buttons=Gtk.ButtonsType.YES_NO,
            text=msg
        )
        confirm.connect("response", self._on_response, origin, (args, kwargs))
        confirm.show_all()
        return True

    def _on_response(self, dialog, response, origin, args):
        (args, kwargs) = args
        if response != Gtk.ResponseType.YES:
            LOGGER.info("User cancelled")
            dialog.destroy()
            self.core.call_all(
                "on_dialog_yes_no_reply", origin, False, *args, **kwargs
            )
            return
        dialog.destroy()
        self.core.call_all(
            "on_dialog_yes_no_reply", origin, True, *args, **kwargs
        )



#!/usr/bin/env python3

import os
import sys

from setuptools import setup, find_packages


quiet = '--quiet' in sys.argv or '-q' in sys.argv

try:
    with open("src/openpaperwork_gtk/_version.py", "r") as file_descriptor:
        version = file_descriptor.read().strip()
        version = version.split(" ")[2][1:-1]
    if not quiet:
        print("OpenPaperwork-gtk version: {}".format(version))
    if "-" in version:
        version = version.split("-")[0]
except FileNotFoundError:
    print("ERROR: _version.py file is missing")
    print("ERROR: Please run 'make version' first")
    sys.exit(1)


setup(
    name="openpaperwork-gtk",
    version=version,
    description=(
        "OpenPaperwork GTK plugins"
    ),
    long_description="""Paperwork is a GUI to make papers searchable.

A bunch of plugins for Paperwork related to GLib and GTK.
    """,
    url=(
        "https://gitlab.gnome.org/World/OpenPaperwork/paperwork/tree/master/"
        "openpaperwork-gtk"
    ),
    download_url=(
        "https://gitlab.gnome.org/World/OpenPaperwork/paperwork/-"
        "/archive/{}/paperwork-{}.tar.gz".format(version, version)
    ),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        ("License :: OSI Approved ::"
         " GNU General Public License v3 or later (GPLv3+)"),
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
    ],
    license="GPLv3+",
    author="Jerome Flesch",
    author_email="jflesch@openpaper.work",
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    zip_safe=(os.name != 'nt'),
    install_requires=[]
)



import logging

try:
    from gi.repository import Gio
    GLIB_AVAILABLE = True
except ImportError:
    GLIB_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.deps


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 75

    def get_interfaces(self):
        return [
            'chkdeps',
            'external_apps',
        ]

    def get_deps(self):
        return []

    def chkdeps(self, out: dict):
        if not GLIB_AVAILABLE:
            out['glib'] = openpaperwork_core.deps.GLIB

    def external_app_open_file(self, file_url):
        LOGGER.info("Opening file '%s' using Gio", file_url)
        if not Gio.AppInfo.launch_default_for_uri(file_url):
            LOGGER.warning("Failed to opening file '%s' using Gio", file_url)
            return None
        return True

    def external_app_open_folder(self, folder_url):
        return self.external_app_open_file(folder_url)
